{
  "type": "record",
  "name": "Metadata",
  "namespace": "pfb",
  "fields": [
    {
      "name": "nodes",
      "type": {
	"type": "array",
	"items": "pfb.Node"
      }
    },
    {
      "name": "misc",
      "type": {
	"type": "map",
	"values": "string"
      }
    }
  ]
}
