{
  "name": "Link",
  "type": "record",
  "namespace": "pfb",
  "fields": [
    {
      "name": "multiplicity",
      "type": {
        "name": "Multiplicity",
        "type": "enum",
        "symbols": [
          "ONE_TO_ONE",
          "ONE_TO_MANY",
          "MANY_TO_ONE",
          "MANY_TO_MANY"
        ]
      }
    },
    {
      "name": "dst",
      "type": "string"
    },
    {
      "name": "name",
      "type": "string"
    }
  ]
}
